use core::{
    alloc::{AllocError, Allocator, Layout},
    borrow::{Borrow, BorrowMut},
    cell::{Ref, RefCell, RefMut},
    mem::MaybeUninit,
    ops::{Deref, DerefMut},
    ptr::NonNull,
};

use alloc::{borrow::ToOwned, boxed::Box, sync::Arc, vec, vec::Vec};
use log::{debug, info};
use uefi::{
    data_types::Align,
    proto::media::{
        file::{Directory as UefiDirectory, File, FileAttribute, FileHandle, FileInfo, FileMode},
        fs::SimpleFileSystem,
    },
    table::boot::ScopedProtocol,
    CStr16, ResultExt, Status,
};

use crate::error::{Error, Result};

pub struct FileSystem<'a> {
    inner: RefCell<ScopedProtocol<'a, SimpleFileSystem>>,
}

pub struct FileInfoIterator {
    dir: Directory,
}

impl FileInfoIterator {
    pub fn new(dir: Directory) -> Self {
        Self { dir }
    }

    pub fn next_entry(&mut self) -> Result<Option<Box<FileInfo>>> {
        let layout = Layout::from_size_align(0, FileInfo::alignment())?;
        let mut ptr = alloc::alloc::Global.allocate(layout)?;

        if let Some(error) = self.dir.read_entry(unsafe { ptr.as_mut() }).err() {
            match error.status() {
                Status::BUFFER_TOO_SMALL => {
                    let layout = Layout::from_size_align(
                        error.data().expect("no size given after buffer_too_small!"),
                        FileInfo::alignment(),
                    )?;
                    let mut ptr = alloc::alloc::Global.allocate(layout)?;

                    let file_info = self
                        .dir
                        .read_entry(unsafe { ptr.as_mut() })
                        .map(|info| {
                            info.map(|info| unsafe { Box::<FileInfo>::from_raw(info as *mut _) })
                                .ok_or_else(|| unsafe {
                                    alloc::alloc::Global.deallocate(ptr.cast(), layout);
                                    AllocError
                                })
                        })
                        .discard_errdata()??;
                    Ok(Some(file_info))
                }
                _ => Err(error).discard_errdata()?,
            }
        } else {
            Ok(None)
        }
    }
}

pub mod dir {
    use core::{
        alloc::{AllocError, Allocator, Layout},
        ops::Deref,
    };

    use alloc::boxed::Box;
    use uefi::{
        data_types::Align,
        proto::media::file::{File, FileAttribute, FileHandle, FileInfo, FileMode},
        ResultExt, Status,
    };

    use crate::Result;

    use super::Directory;

    pub struct DirectoryEntry {
        dir: Directory,
        file_info: Box<FileInfo>,
    }

    impl core::fmt::Debug for DirectoryEntry {
        fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
            f.debug_struct("DirectoryEntry")
                .field("file_info", &self.file_info)
                .finish()
        }
    }

    impl DirectoryEntry {
        pub fn open(
            &self,
            open_mode: FileMode,
            attribs: Option<FileAttribute>,
        ) -> Result<FileHandle> {
            let handle = self.dir.get_mut().open(
                self.file_info.file_name(),
                open_mode,
                attribs.unwrap_or(FileAttribute::empty()),
            )?;

            Ok(handle)
        }
    }

    impl AsRef<FileInfo> for DirectoryEntry {
        fn as_ref(&self) -> &FileInfo {
            &self.file_info
        }
    }

    impl Deref for DirectoryEntry {
        type Target = FileInfo;

        fn deref(&self) -> &Self::Target {
            &self.file_info
        }
    }

    impl Directory {
        pub fn next_entry(&mut self) -> Result<Option<Box<FileInfo>>> {
            let layout = Layout::from_size_align(0, FileInfo::alignment())?;
            let mut ptr = alloc::alloc::Global.allocate(layout)?;

            if let Some(error) = self.read_entry(unsafe { ptr.as_mut() }).err() {
                match error.status() {
                    Status::BUFFER_TOO_SMALL => {
                        let layout = Layout::from_size_align(
                            error.data().expect("no size given after buffer_too_small!"),
                            FileInfo::alignment(),
                        )?;
                        let mut ptr = alloc::alloc::Global.allocate(layout)?;

                        let file_info = self
                            .read_entry(unsafe { ptr.as_mut() })
                            .map(|info| {
                                info.map(|info| unsafe {
                                    Box::<FileInfo>::from_raw(info as *mut _)
                                })
                                .ok_or_else(|| unsafe {
                                    alloc::alloc::Global.deallocate(ptr.cast(), layout);
                                    AllocError
                                })
                            })
                            .discard_errdata()??;
                        Ok(Some(file_info))
                    }
                    _ => Err(error).discard_errdata()?,
                }
            } else {
                Ok(None)
            }
        }
    }

    impl Iterator for Directory {
        type Item = DirectoryEntry;

        fn next(&mut self) -> Option<Self::Item> {
            self.next_entry()
                .ok()
                .flatten()
                .map(|file_info| DirectoryEntry {
                    dir: self.clone(),
                    file_info,
                })
        }
    }
}

impl Iterator for FileInfoIterator {
    type Item = Box<FileInfo>;

    fn next(&mut self) -> Option<Self::Item> {
        self.next_entry().ok().flatten()
    }
}

#[repr(transparent)]
#[derive(Clone)]
pub struct Directory {
    dir: Arc<RefCell<UefiDirectory>>,
}

impl Directory {
    pub fn get(&self) -> Ref<'_, UefiDirectory> {
        RefCell::borrow(&self.dir)
    }

    pub fn get_mut<'a>(&'a self) -> RefMut<'a, UefiDirectory> {
        RefCell::borrow_mut(&self.dir)
    }

    pub fn open(
        &self,
        filename: &CStr16,
        open_mode: FileMode,
        attributes: FileAttribute,
    ) -> uefi::Result<FileHandle> {
        self.get_mut().open(filename, open_mode, attributes)
    }

    pub fn read_entry<'buf>(
        &mut self,
        buffer: &'buf mut [u8],
    ) -> uefi::Result<Option<&'buf mut FileInfo>, Option<usize>> {
        self.get_mut().read_entry(buffer)
    }
}

impl From<UefiDirectory> for Directory {
    fn from(value: UefiDirectory) -> Self {
        Self {
            dir: Arc::new(RefCell::new(value)),
        }
    }
}

impl<'a> FileSystem<'a> {
    pub fn new(inner: ScopedProtocol<'a, SimpleFileSystem>) -> Self {
        Self {
            inner: RefCell::new(inner),
        }
    }

    pub fn open(
        &self,
        path: &str,
        open_mode: FileMode,
        attribs: Option<FileAttribute>,
    ) -> Result<FileHandle> {
        let mut buf = vec![0; path.len() + 1];
        let cstr = CStr16::from_str_with_buf(path, &mut buf)?;
        Ok(self.root_dir()?.get_mut().open(
            cstr,
            open_mode,
            attribs.unwrap_or(FileAttribute::empty()),
        )?)
    }

    pub fn root_dir(&self) -> Result<Directory> {
        Ok(self.inner.borrow_mut().open_volume()?.into())
    }
}
