#![feature(abi_efiapi)]
#![feature(allocator_api)]
#![feature(error_in_core)]
#![feature(alloc_error_handler)]
#![feature(negative_impls)]
#![feature(int_roundings)]
#![feature(core_intrinsics)]
#![feature(iter_intersperse)]
#![no_std]
#![no_main]

use alloc::{collections::BTreeMap, string::String, vec};
use graphics::GraphicsOutput;
use log::info;
use serde::Deserialize;
use uefi::{
    prelude::*,
    proto::{
        console::gop,
        media::file::{File, FileAttribute, FileMode},
    },
    table::{
        boot::{OpenProtocolAttributes, OpenProtocolParams, SearchType},
        Boot, SystemTable,
    },
    Identify, Status,
};

extern crate alloc;

mod context;
mod error;
mod fpu;
mod fs;
mod graphics;
mod input;

use error::{Error, Result};

use crate::fpu::enable_fpu;

#[derive(Debug, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct Config {
    #[serde(rename = "FREELOADER")]
    freeloader: FreeLoaderConfig,
}

#[derive(Debug, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct FreeLoaderConfig {
    timeout: u64,
    #[serde(rename = "DefaultOS")]
    default_os: String,
    font_path: String,
}

#[derive(Debug)]
pub struct Context {
    handle: Handle,
    st: SystemTable<Boot>,
}

impl Context {
    pub fn new(handle: Handle, mut st: SystemTable<Boot>) -> uefi::Result<Self> {
        uefi_services::init(&mut st)?;
        Ok(Self { handle, st })
    }

    pub fn fs(&self) -> Result<fs::FileSystem> {
        let file_system = self.st.boot_services().get_image_file_system(self.handle)?;

        Ok(fs::FileSystem::new(file_system))
    }

    pub fn system_table(&self) -> &SystemTable<Boot> {
        &self.st
    }

    pub fn handle(&self) -> Handle {
        self.handle
    }

    pub fn system_table_mut(&mut self) -> &mut SystemTable<Boot> {
        &mut self.st
    }

    pub fn boot_services(&self) -> &BootServices {
        self.st.boot_services()
    }

    pub fn get_graphics_output(&self) -> Result<GraphicsOutput> {
        let handles = self
            .boot_services()
            .locate_handle_buffer(SearchType::ByProtocol(&gop::GraphicsOutput::GUID))?;

        let gops = handles.handles().iter().filter_map(|handle| unsafe {
            self.boot_services()
                .open_protocol::<gop::GraphicsOutput>(
                    OpenProtocolParams {
                        handle: handle.clone(),
                        agent: self.handle,
                        controller: None,
                    },
                    OpenProtocolAttributes::GetProtocol,
                )
                .ok()
        });

        let (_, (gop, mode)) = gops
            .map(|gop| {
                let (pixel_count, mode) = gop
                    .modes()
                    .map(|mode| {
                        let (x, y) = mode.info().resolution();
                        (x * y, mode)
                    })
                    .collect::<BTreeMap<_, _>>()
                    .pop_last()
                    .expect("gop has 0 modes !?");

                (pixel_count, (gop, mode))
            })
            .collect::<BTreeMap<_, _>>()
            .pop_last()
            .expect("no GOP found!");

        GraphicsOutput::from_gop_with_mode(gop, mode)
    }

    pub fn run(mut self) -> Result<Status> {
        info!("Hello, UEFI!");
        self.system_table_mut().stdin().reset(false)?;

        enable_fpu().expect("no FPU/SSE/SSE2!");

        let fs = self.fs()?;

        for entry in fs.root_dir()?.into_iter() {
            info!("{:#?}", entry);
            if !entry.attribute().contains(FileAttribute::DIRECTORY) {
                let mut file = entry
                    .open(FileMode::Read, None)?
                    .into_regular_file()
                    .unwrap();

                let mut buf = vec![0u8; entry.file_size() as usize];
                _ = file.read(&mut buf).unwrap();
                let s = String::from_utf8_lossy(&buf);
                info!("{:?}", s);
            }
        }

        let mut display = self.get_graphics_output()?;
        let (width, height) = display.resolution();

        info!("creating terminal");

        loop {}

        Ok(Status::SUCCESS)
    }
}

#[entry]
fn efi_entry(handle: Handle, system_table: SystemTable<Boot>) -> Status {
    let ctx = Context::new(handle, system_table).unwrap();
    ctx.run().unwrap()
}
